#!/usr/bin/python
#
#  Ezca: Epics EZCA interface in Python
#
#  Copyright (c) 2000--2002  Matt Newville, The University of Chicago
#
#  email:       newville@cars.uchicago.edu
#
#  versions
#    1.00   initial use, debugging of connect, start/end group
#    1.10   added optional use of Numeric data type, switched
#           build to use distutils, and updated docs
#    1.11   added "type='char'" option to caget to try to force
#           data to be returned as a string.  For many record types 
#           (mbbi, mbbo), this has the added benefit of retrieving
#           the string value for the field.
#           also add "use_char=1" option to initialization, which
#           will set caget to use character strings when appropriate.
#    1.12   improved support for long character arrays (byte-arrays)
#           from Epics string waveform records.  these allow strings
#           longer than 40 characters to be used in Epics records.
#    1.13   changed installation to include Geoff Savage's ca extension,
#           and build windows extensions.
#
#   last update: 17-Aug-2002
#
from ezcac import *
import types, string

class Ezca:
    """Ezca: Epics EZCA interface in Python

 example:
    import Ezca
    ez = Ezca.Ezca(timeout = 0.001, retry=20)
    x = ez.caget('XX:m1.VAL')

 The Ezca class provides an interface to the EPICS Ezca library
 from Python. The supported functions are:

    caput(pv,val)      :  set a PV value
    caget(pv)          :  get a PV value
    getpv(pv, options) :  enhanced PV get (see doc below)
    set_debug()        :  set debug flag
    show_setup         :  show configuration of Ezca instance.
    get_status         :  return last Ezca exit status
    connect            :  connect to a pv, saving count,type in local cache
    use_numeric        :  set the use of Numeric Python arrays for array types
    GetCountAndType    :  return PV count and _python_ type
    GetNelem           :  return PV count
    GetStatus          :  Get time, status, and severity for a PV
    GetUnits           :  return units of a PV
    GetPrecision       :  return precision of a PV
    SetMonitor         :  set a PV monitor
    ClearMonitor       :  unset a PV monitor
    CheckMonitor       :  check status of a PV monitor
    StartGroup         :  start a group of efficient Ezca calls
    EndGroup           :  end a group of efficient Ezca calls
    delay              :  set an Ezca time delay

 instance options for:
    import Ezca
    ez = Ezca.Ezca(option=value, ....)

    Several optional arguments can be given to affect the 
    behavior of the Ezca instance:

     option          meaning                                     default
       timeout        ca timeout (in sec)                          0.001
       retry          ca retry count                               300
       debug          debug flag (1: on, 0: off)                   0
       trace          flag to turn ezcaTraceOn on/off              0
       autoerror      flag to turn ezcaAutoError on/off            0

       use_char       flag to usw string output where appropriate  0
       use_numeric    flag to use Numeric arrays for array data    0

    Both 'use_char' and 'use_numeric' are recommended!
"""

    version   = '1.13'

    CaTypeMap = ( (1, "char"),    (2, "int"),
                  (4, "float"),   (2, "int"),
                  (0, "byte"),     (3, "long"),
                  (5, "double") )

    
    def __init__(self,**args):
        self.PVdata    =  {}     # storage of local cache of count/type
        self.retry     =  300
        self.timeout   =  0.001
        self.debug     =  0
        self.trace     =  0
        self.autoerror =  0
        self.status    =  0
        self.in_group  =  0
        self.use_char  =  0
        self.numeric   =  0
        for k in args.keys():
            if (k == 'debug'):        self.debug      = args[k]
            if (k == 'trace'):        self.trace      = args[k]
            if (k == 'retry'):        self.retry      = args[k]
            if (k == 'autoerror'):    self.autoerror  = args[k]
            if (k == 'timeout'):      self.timeout    = args[k]
            if (k == 'use_numeric'):  self.numeric    = args[k]
            if (k == 'use_char'):     self.use_char   = args[k]
            

        i = ezcaSetRetryCount(self.retry)
        i = ezcaSetTimeout(self.timeout)
        i = ezcaTraceOff()
        i = ezcaAutoErrorMessageOff()
        if (self.autoerror != 0): i = ezcaAutoErrorMessageOn()
        if (self.trace != 0):     i = ezcaTraceOn()
        if (self.numeric != 0):   i = self.use_numeric()


    def use_numeric(self,val=1):
        """set the use of Numeric Python arrays for array types
           ez.use_numeric()          turn on Numeric arrays
           ez.use_numeric(val=0)     turn off Numeric arrays
        """
        self.numeric = val
        if (self.numeric == 1):
            try:
                import Numeric
                self.get_num_array = Numeric.array
                self.numeric = 1
            except:
                print " Warning: cannot use the Numeric Module!!"
                self.numeric = 0

    def set_debug(self,d=1):
        """set debug status:
           ezca.set_debug(1)   turn debugging on
           ezca.set_debug()    turn debugging on
           ezca.set_debug(0)   turn debugging off
        """
        self.debug = d
        return self.debug

    def get_status(self):
        """get Ezca status """
        return self.status

    def show_setup(self):
        "print current setup to standard output"
        print " status      =", self.status
        print " retry count =", self.retry
        print " timeout     =", self.timeout
        print " debug       =", self.debug
        print " trace       =", self.trace
        print " autoerror   =", self.autoerror
        print " use_numeric =", self.use_numeric
        print " use_char    =", self.use_char
        print " Known PVs with types and Counts:"
        print "        PV     (Count,  Ezca Type,  C/Python Type)"
        for i in self.PVdata.keys(): print i[:3],  self.PVdata[i]
        return self.status

    def caget(self,pv,fromcache=None,type=None):
        """get a Process Variable:
 x = ezca.caget(pv)
 options:
    fromcache=1    do _not_ do a real caget, but get
                   the last read value, as after a set
                   of StartGroup/EndGroup.
    type='char'    try to force output as string (useful for
                   bo//bi/mbbi/mbbo output)

    """
        if (self.debug == 1): print " Caget: ", pv
        try:
            (cnt, eztype, pytype, p) = self.connect(pv)
        except:
            return -1
        out = 0 ; stat = 0
        request_type = type
        if (request_type==None):
            request_type = pytype
            if ((self.use_char == 1) and (pytype=='int')):
                request_type='char'

        if (pytype=='byte'): request_type='char'

        if (cnt != 0):
            if (request_type == 'char'):
                str   = " "*128 ; len = 0 ; is_char =1 ; out = " "
                try:
                    if (cnt == 1):
                        stat  = ezca_GetChar(pv, str, p)
                        if (stat == 0):  len = ptrvalue(p)
                        if (len < 1): len = 1
                        out = str[0:len]
                    else:
                        stat  = ezcaGet(pv, eztype, cnt, p)                    
                        out   = ptrvalue(p, 0, request_type)
                except TypeError:
                    stat  = ezcaGet(pv, eztype, cnt, p)                    
                    out   = repr(ptrvalue(p, 0, pytype))
                if (self.debug == 1): print " Caget output = ", out, cnt
            else:
                if (fromcache==None):  stat   = ezcaGet(pv, eztype, cnt, p)
                if (stat == 0):
                    if (cnt == 1):
                        out = ptrvalue(p, 0, pytype)
                    else:
                        out = []
                        for i in range(cnt):
                            out.append( ptrvalue(p,i, pytype) )
                        if (self.numeric == 1):
                            out = self.get_num_array(out)
        self.status = stat
        if (self.status != 0): out = None
        return out
    ##
    def caput(self, pv, data):
        """set a Process Variable:
          x = ezca.caput(pv,val)"""
        # given a pv name, return ezcaGet status and fill in
        # second argument with data array
        npts = 0
        (c, t, v, p) = self.connect(pv)
        if (self.debug == 1):
            print " caput:  pv: %s  ||type:%s\n  data: %s" %(pv,type(data),data)

        if (isinstance(data,types.StringType) and (v=='double')):
            data = float(data)
        elif (isinstance(data,types.StringType) and (v == 'int')):
            data = int(data)
        stat   = -1
        if (v == "char"):
            stat  = ezca_PutChar(pv, data)
        elif ((v == "byte") and (c > 0)):
            p1  = ptrcast(p,'char *')
            if (len(data)>c): data = data[:c]
            ptrset(p1,data)
            stat  = ezcaPutOldCa(pv, 0, c, p1)
            # ptrfree(p1)
        else:
            if (isinstance(data,types.ListType)):
                if (self.debug == 1):  print " ARRAY : npts = " , c
                if (c != 0):
                    if (self.numeric == 1):   data = list(data)
                    for i in range(c):  ptrset(p, data[i], i)
                    stat = ezcaPutOldCa(pv, t, c, p)
            else:
                ptrset(p, data)
                if (self.debug == 1):
                    print " data : ", data, " type = ", type(data)
                stat = ezcaPutOldCa(pv, t, c, p)
            # print "CAPUT data : ", pv, " val = ", data,  p, stat
        if (self.debug == 1):  print " Caput status : ", stat
        self.status = stat
        return stat

    def connect(self,pv):
        """initiate a connection to a PV,
        internally storing useful information about the PV
        such as count, type, rtype, and allocating a  pointer
        for the data.
           res = ezca.connect(pv)
        returns a tuple of count, ezcatype, python_type, and pointer
        """
        # given a pv name, return a list of (count,  ezcatype, pointer)
        # where p is the python pointer
        if (self.PVdata.has_key(pv) == 0):
            count = 1;  ezcatype = 2; pytype = "int"; p=0
            p_t = ptrcreate("int")
            p_c = ptrcreate("int")
            s   = ezca_GetCountAndType(pv, p_c, p_t)
            if (s == 0):
                type   = ptrvalue(p_t)
                count  = ptrvalue(p_c)
                if (type > 6): type = 2
                (ezcatype, pytype) = self.CaTypeMap[type]
                if (pytype in ('char','int','byte')):
                    p = ptrcreate("int",0,count)
                else:
                    p = ptrcreate(pytype,0,count)
            ptrfree(p_c)
            ptrfree(p_t)
            #  save valuse for count, ezcatype and pytype in hashes
            self.PVdata[pv] = (count,ezcatype,pytype,p)

        ret = self.PVdata[pv]
        if (self.debug == 1):
            print " Connect: ", pv, \
                  " : (count, ezcatype, pytype, pointer) ", ret
        return ret

    def GetCountAndType(self,pv):
        """Get Count and Type (python type!) of a PV:
            (count,type) = ezca.GetNelem(pv)"""
        (c, t, v, p) = self.connect(pv)
        return (c,v)

    def GetNelem(self,pv):
        """Get Number of Elements (count) of a PV:
           count = ezca.GetNelem(pv)"""
        (c, t, v, p) = self.connect(pv)
        return c

    def GetStatus(self,pv):
        """Get time, status, and severity for a PV:
        (time, status, severity) = ezca.GetStatus(pv)"""
        (p_time, p_stat, p_sev) = (0,0,0)
        (time, sev)             = (-1,0)
        p_time = ptrcreate("int")
        p_stat = ptrcreate("int")
        p_sev  = ptrcreate("int")
        stat   = ezca_GetStatus_(pv,p_time,p_stat,p_sev)
        if (stat != 0):
            time         = ptrvalue(p_time)
            self.status  = ptrvalue(p_stat)
            sev          = ptrvalue(p_sev)
        return (time,stat,sev);

    def GetUnits(self,pv):
        """Get Units (as a character string) of a PV:
           units = ezca.GetUnits(pv)"""
        (s, len, units, str) = (0,0,"","")
        (c, t, v, p) = self.connect(pv)
        if (c != 0):
            import string
            str   = " "*16
            p_len = ptrcreate("int")
            s     = ezca_GetUnits_(pv,str,p_len)
            len   = ptrvalue(p_len)
            units = str[0:len]
        return units

    def GetPrecision(self,pv):
        """Get Precision of a PV:
           prec = ezca.GetPrecision(pv)"""
        (s, p, prec) = (0,0,0)
        (c, t, v, p) = self.connect(pv)
        if (c != 0):
            p    = ptrcreate("short")
            s    = ezcaGetPrecision(pv,p)
            prec = ptrvalue(p)
        return prec

    def SetMonitor(self,pv):
        """Set a Monitor on a PV:
           status  = ezca.SetMonitor(pv)"""
        (c, t, v, p) = self.connect(pv)
        s = 0
        if (c != 0):  s = ezcaSetMonitor(pv, t)
        return s

    def ClearMonitor(self,pv):
        """Clear a Monitor on a PV:
           status  = ezca.ClearMonitor(pv)"""
        (c, t, v, p) = self.connect(pv)
        s = 0
        if (c != 0): s = ezcaClearMonitor(pv, t)
        return s

    def CheckMonitor(self,pv):
        """Check the Status of Monitored PV:
           status  = ezca.CheckMonitor(pv)"""
        (c, t, v, p) = self.connect(pv)
        s = 0
        if (c != 0): s = ezcaNewMonitorValue(pv, t)
        return s

    def delay(self,time):
        """Do an Ezca time delay:
        status  = ezca.delay(time)"""
        ret = ezcaDelay(time)
        return ret

    def StartGroup(self):
        """Start a 'Group' for efficient get/puts:
        status  = ezca.StartGroup()"""
        self.in_group = 1
        ret = ezcaStartGroup()
        return ret

    def EndGroup(self):
        """End a 'Group' for efficient get/puts:
        status  = ezca.EndGroup()"""
        self.in_group = 0
        ret = ezcaEndGroup()
        return ret

    def getpv(self,pv,type='scalar',format=None,descpv=None):
        """Enhanced caget of a PV:
        (val, desc) = ezca.getpv(pv, options)

        where the options controlling what the output value and
        and description are, and how they are formatted.

        In general, a PV is assumed to have a '.VAL' (which is
        optional), and the corresponding '.VAL'   and '.DESC'
        fields will be found and passed back:
        val,desc = ezca.getpv('XXX:m1')
        will return the values of XXX:m1.VAL and XXX:m1.DESC

        option     meaning
        type       PV type (full, bo, mbbi)
                   'bo':   PV is a binary out, lookup '.ZNAM'
                           or '.ONAM' field as appropriate
                           for the output VALUE
                   'mbbi': PV is a mbbi, lookup '.ZRST', ...
                           as appropriate for the output VALUE
                   'full': Do _NOT_ append '.VAL' and '.DESC'
                           to PV.
                   'yes/no': translate 1/0 to 'Yes/No'
        desc       PV to use to lookup description field (instead
                   of using the '.DESC' field.
        format     C/Python format string to use to format output """
        pv = string.strip(pv)
        (pvname, pvpre) = (pv+'.VAL',pv)
        if (pv[-4:] == '.VAL'):  (pvname,pvpre) = (pv, pv[:-4])
        if (pv[-5:] == '.DESC'): (pvpre,descpv) = (pv[:-5],None)        
        if (type == "full"): pvname = pv
        val  = self.caget(pv)
        if (self.status !=0):  val = self.caget(pvname)
        if (descpv == None): descpv  = pvpre + '.DESC'
        desc = self.caget(descpv)
        if ((self.status !=0) or
            (desc==None) or (desc == "")): desc = pv
        if (type == "mbbi"):
            try:
                dx  = (".ZRST", ".ONST",".TWST", ".THST",".FRST", ".FVST")[val]
                val = self.caget(pv+dx)
            except:
                val = None
        elif (type == "bo"):
            try:
                dx  = (".ZNAM", ".ONAM")[val]
                val = self.caget(pv+dx)
            except:
                val = None
        elif (format != None):
            val = format % val
        npts = self.GetNelem(pv)
        if ((npts > 1) and (type != "full")):
            val = "Array (%i  points)" %npts
        if (type == "yes/no"):
            res = 'No'
            if (val): res = 'Yes'
            val = res
        return (val,desc)

if (__name__ == '__main__'):
    import sys
    ez  = Ezca(timeout=0.0005, retrycount=300,use_char=1)
    print "Type Process Variable Names or 'quit' to end"
    while 1:
        try:
            pv  = raw_input("PV >")
            if (pv == 'quit'): break
            if (pv != ''):
                val = ez.caget(pv)
                stat= ez.get_status()
                if (stat == 0):
                    print pv, " = ", val
                else:
                    print "Error status = ", stat, " for " , pv
        except:
            break

